/* LatVis
 * Copyright (C) 2002 Jan Outrata <outrataj@phoenix.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "layerdiagram.h"
#include "orderdata_array.h"
#include "orderdata_list.h"
#include "orderdata_tree.h"
#include <stdio.h>

bool
LayerDiagram::isEdge(Vertex_e *vx, Vertex_e *vy)
{
  if ((vx->element != NoElement) && (vy->element != NoElement)) {
    // vx, vy proper or mixed
    if ((vx->dummy != NoElement) && (vy->dummy != NoElement))
      // vx, vy mixed
      return (dummy_chains_data && 
	      (!dummy_chains_data->less(vx->dummy, vy->dummy, true)) &&
	      data->less(vx->element, vy->element, false));
    return data->less(vx->element, vy->element, false);
  }
  if (dummy_chains_data && (vx->dummy != NoElement) && 
      (vy->dummy != NoElement))
    // vx, vy dummy or mixed
    return dummy_chains_data->less(vx->dummy, vy->dummy, false);
  return false;
}

void
LayerDiagram::longestPathLayering()
{
  /* Place verteces of all maximal elements to layer 0.
     Place the vertex of the element to layer n down, where n is
     the length of the longest path from the element to the maximal
     element.
  */

  set<ElementId> elems, first_elems;
  set<ElementId>::iterator iter_s1, iter_s2;
  Vertex_e *vert;
  list<Vertex_e *> *layer;
  int n, l, layer_count;
  list<set<ElementId> > paths;
  list<set<ElementId> >::iterator iter_l;
  list<list<Vertex_e *> *>::iterator iter_ll;
  
  data->getElements(elems);
  if (elems.empty()) return;
  if (generationParams->from_maximal) 
    data->maximalElements(first_elems, elems);
  else data->minimalElements(first_elems, elems);
  if (first_elems.empty()) return;
  layer = new list<Vertex_e *>;
  if (!layer) return;
  for (iter_s1 = first_elems.begin(); iter_s1 != first_elems.end();
       iter_s1++) {
    vert = new Vertex_e;
    vert->element = *iter_s1; vert->dummy = NoElement;
    layer->push_back(vert);
    elems.erase(*iter_s1);
  }
  layers.push_back(layer);
  for (iter_s1 = elems.begin(); iter_s1 != elems.end(); iter_s1++) {
    n = 0;
    for (iter_s2 = first_elems.begin(); iter_s2 != first_elems.end();
	 iter_s2++) {
      paths.clear();
      if (generationParams->from_maximal) 
	data->getPaths(paths, *iter_s1, *iter_s2);
      else data->getPaths(paths, *iter_s2, *iter_s1);
      for (iter_l = paths.begin(); iter_l != paths.end(); iter_l++) {
	l = (*iter_l).size() - 1; // count edges
	if (l > n) n = l;
      }
    }
    layer_count = layers.size() - 1;
    for (l = 0; l < (n - layer_count); l++) {
      layer = new list<Vertex_e *>;
      layers.push_back(layer);
    }
    vert = new Vertex_e;
    vert->element = *iter_s1; vert->dummy = NoElement;
    for (l = 0, iter_ll = layers.begin(); l < n; l++, iter_ll++);
    (*iter_ll)->push_back(vert);
  }
}

bool
LayerDiagram::lessNumberSet(set<cg_number_t> s1, set<cg_number_t> s2)
{
  set<cg_number_t>::iterator iter_s1, iter_s2;

  if (s1.empty() && (!s2.empty())) return true;
  if (s2.empty()) return false; // also satisfies both sets empty
  iter_s1 = s1.end(); iter_s1--; iter_s2 = s2.end(); iter_s2--;
  if (*iter_s1 < *iter_s2) return true;
  if (*iter_s2 < *iter_s1) return false;
  s1.erase(iter_s1); s2.erase(iter_s2);
  return lessNumberSet(s1, s2);
}

void
LayerDiagram::CoffmanGrahamLayering(unsigned int w)
{
  /* Assign 0 to the vertex of minimal element (arbitrarily chosen).
     Next number assign to the vertex of the element, which doesn't
     have a number assigned, all verteces of lower elements have number
     assigned and the set of numbers assigned to verteces of lower
     elements is minimal (in the sense of the set order defined by
     method lessNumberSet).
     Place the vertex of element, which is not placed, all verteces of
     upper elements are placed. If there are more such verteces, place
     the one with biggest number. If there is no one or the layer is full
     (max. width w), go and process the lower layer.
  */

  set<ElementId> elems, elems1;
  cg_NumberElement nelem;
  set<cg_NumberElement> nelems;
  set<cg_NumberElement>::iterator iter_sn;
  cg_number_t number = 0;
  set<ElementId>::iterator iter_s, iter_s1;
  cg_ElemNumberSet ens;
  const cg_ElemNumberSet *min_ens;
  set<cg_ElemNumberSet> set_numbers;
  set<cg_ElemNumberSet>::iterator iter_sns;
  Vertex_e *vert;
  list<Vertex_e *> *layer;
  list<list<Vertex_e *> *>::iterator iter_ll;
  list<Vertex_e *>::iterator iter_l;

  if (!w) return;
  data->getElements(elems);
  if (elems.empty()) return;
  if (generationParams->from_maximal) data->minimalElements(elems1, elems);
  else data->maximalElements(elems1, elems);
  if (elems1.empty()) return;
  nelem.element = *(elems1.begin()); nelem.number = number++;
  nelems.insert(nelem); elems.erase(nelem.element);
  while (elems.size()) {
    set_numbers.clear();
    for (iter_s = elems.begin(); iter_s != elems.end(); iter_s++) {
      ens.element = *iter_s; ens.numbers.clear();
      elems1.clear();
      if (generationParams->from_maximal) 
	data->lowerElements(elems1, *iter_s);
      else data->upperElements(elems1, *iter_s);
      for (iter_s1 = elems1.begin(); iter_s1 != elems1.end(); iter_s1++) {
	for (iter_sn = nelems.begin(); iter_sn != nelems.end(); iter_sn++)
	  if ((*iter_sn).element == *iter_s1) {
	    ens.numbers.insert((*iter_sn).number);
	    break;
	  }
	if (iter_sn == nelems.end()) break; 
	// lower element doesn't have a number
      }
      if (iter_s1 != elems1.end()) continue;
      set_numbers.insert(ens);
    }
    min_ens = &(*(set_numbers.begin()));
    for (iter_sns = set_numbers.begin(), iter_sns++;
	 iter_sns != set_numbers.end(); iter_sns++)
      if (lessNumberSet((*iter_sns).numbers, min_ens->numbers))
	min_ens = &(*iter_sns);
    nelem.element = min_ens->element; nelem.number = number++;
    nelems.insert(nelem); elems.erase(nelem.element);
  }
  layer = new list<Vertex_e *>;
  if (!layer) return;
  while (nelems.size()) {
    for (iter_sn = nelems.begin(); iter_sn != nelems.end(); iter_sn++) {
      elems1.clear();
      if (generationParams->from_maximal) 
	data->upperElements(elems1, (*iter_sn).element);
      else data->lowerElements(elems1, (*iter_sn).element);
      for (iter_s1 = elems1.begin(); iter_s1 != elems1.end(); iter_s1++) {
	for (iter_ll = layers.begin(); iter_ll != layers.end(); iter_ll++) {
	  for (iter_l = (*iter_ll)->begin(); iter_l != (*iter_ll)->end();
	       iter_l++)
	    if ((*iter_l)->element == *iter_s1) break;
	  if (iter_l != (*iter_ll)->end()) break;
	}
	if (iter_ll == layers.end()) break;
	// upper element isn't placed
      }
      if (iter_s1 != elems1.end()) continue;
      break;
    }
    if ((iter_sn == nelems.end()) || (layer->size() == w)) {
      layers.push_back(layer);
      layer = new list<Vertex_e *>;
    }
    else { 
      vert = new Vertex_e;
      vert->element = (*iter_sn).element; vert->dummy = NoElement;
      layer->push_back(vert);
      nelems.erase(iter_sn);
    }
  }
  layers.push_back(layer);
}

void
LayerDiagram::insertDummyVerteces()
{
  /* For all edged (u, v) longer than 1 insert dummy verteces into
     the layer i (l(u) < i < l(v)) at the x-coordinate
     ((i - l(u)) / (l(v) - l(u))) * (x(v) - x(u)) + x(u), where
     l(u), resp. l(v), is the layer in which is the vertex u, resp. v,
     and x(u), resp. x(v), is the x-coordinate of the vertex u,
     resp. v, in that layer.
  */

  set<ElementId> elems;
  set<ElementId>::iterator iter_s1, iter_s2;
  list<list<Vertex_e *> *>::iterator iter_ll;
  list<Vertex_e *>::iterator iter_l, iter_l_dest;
  int y, y1, y2;
  Coord_e_t x, x1, x2;
  int c;
  Vertex_e *vert, *vert_old = NULL;
  ElementId id = 1;

  data->getElements(elems);
  if (elems.empty()) return;
  for (iter_s1 = elems.begin(); iter_s1 != elems.end(); iter_s1++) {
    for (iter_s2 = elems.begin(); iter_s2 != elems.end(); iter_s2++) {
      if (*iter_s1 == *iter_s2) continue;
      if (!data->less(*iter_s1, *iter_s2, false)) continue;
      for (iter_ll = layers.begin(), y = 0, c = 0; iter_ll != layers.end();
	   iter_ll++, y++) {
	for (iter_l = (*iter_ll)->begin(); iter_l != (*iter_ll)->end();
	     iter_l++) {
	  if ((*iter_l)->element == *iter_s1) { 
	    x1 = (*iter_l)->x; y1 = y; c++;
	  }
	  else if ((*iter_l)->element == *iter_s2) { 
	    x2 = (*iter_l)->x; y2 = y; c++;
	  }
	  if (c == 2) break;
	}
	if (c == 2) break;
      }
      if (!generationParams->from_maximal) { y = y1; y1 = y2; y2 = y; }
      if ((y1 - y2) <= 1) continue;
      vert_old = NULL;
      for (iter_ll = layers.begin(), y = 0, c = 0; iter_ll != layers.end();
	   iter_ll++, y++) {
	if ((y < y2) || (y > y1)) continue;
	iter_l_dest = (*iter_ll)->end();
	vert = NULL;
	x = (((x1 - x2) * (y - y2)) / (y1 - y2)) + x2;
	for (iter_l = (*iter_ll)->begin(); iter_l != (*iter_ll)->end();
	     iter_l++) {
	  if (((*iter_l)->element == *iter_s1) || 
	      ((*iter_l)->element == *iter_s2)) {
	    vert = *iter_l; c++; break;
	  }
	  else if ((y > y2) && (y < y1) && ((*iter_l)->x >= x)) break;
	}
	if (!vert) {
	  if ((iter_l != (*iter_ll)->end()) && (*iter_l)->x == x)
	    (x < 0) ? x -= 1 : x += 1;
	  vert = new Vertex_e;
	  vert->element = vert->dummy = NoElement;
	  if (iter_l != (*iter_ll)->begin()) iter_l--;
	  vert->x = x; vert->y = (*iter_l)->y;
	  iter_l_dest = iter_l;
	}
	if (vert->dummy == NoElement) {
	  vert->dummy = id; dummy_chains_data->newElement(id); id++;
	}
	if (vert_old) {
	  if (generationParams->from_maximal) 
	    dummy_chains_data->orderElements(vert->dummy, vert_old->dummy);
	  else 
	    dummy_chains_data->orderElements(vert_old->dummy, vert->dummy);
	}
	vert_old = vert;
	if (iter_l_dest != (*iter_ll)->end())
	  (*iter_ll)->insert(++iter_l_dest, vert);
	if (c == 2) break;
      }
    }
  }
}

void
LayerDiagram::countCrossingNumbers(list<Vertex_e *> *l1, 
				   list<Vertex_e *> *l2)
{
  /* Crossing number cuv for verteces u, v in layer l2,
     where x2(u) < x2(v), is the number of edge pairs (u, w), (v, z),
     where w, z are verteces in layer l1 with x1(z) < x1(w).
     x1, resp. x2 is the position of vertex in layer l1, resp. l2.
     cuu = 0 for all verteces u in layer l2.
  */

  int n, u, v;
  list<Vertex_e *>::iterator iter_l1_1, iter_l1_2, iter_l2_1, iter_l2_2;

  n = l2->size();
  if (crossing_numbers) {
    for (u = 0; crossing_numbers[u]; u++) delete [] crossing_numbers[u];
    delete [] crossing_numbers; crossing_numbers = NULL;
  }
  crossing_numbers = new crossing_number_t * [n + 1];
  if (!crossing_numbers) return;
  for (u = 0; u < n; u++) {
    crossing_numbers[u] = NULL;
    crossing_numbers[u] = new crossing_number_t [n];
    if (!crossing_numbers[u]) {
      for (u--; u >= 0; u--) delete crossing_numbers[u];
      delete crossing_numbers;
      return;
    }
  }
  crossing_numbers[u] = NULL;
  for (iter_l2_1 = l2->begin(), u = 0; iter_l2_1 != l2->end();
       iter_l2_1++, u++)
    for (iter_l2_2 = l2->begin(), v = 0; iter_l2_2 != l2->end();
	 iter_l2_2++, v++) {
      crossing_numbers[u][v] = 0;
      if (u == v) continue;
      for (iter_l1_1 = l1->begin(); iter_l1_1 != l1->end(); iter_l1_1++)
	for (iter_l1_2 = iter_l1_1, iter_l1_2++; iter_l1_2 != l1->end();
	     iter_l1_2++)
	  if (generationParams->from_maximal) {
	    if ((isEdge(*iter_l2_2, *iter_l1_1)) && 
		(isEdge(*iter_l2_1, *iter_l1_2))) (crossing_numbers[u][v])++;
	  }
	  else
	    if ((isEdge(*iter_l1_1, *iter_l2_2)) && 
		(isEdge(*iter_l1_2, *iter_l2_1))) (crossing_numbers[u][v])++;
    }
}

void
LayerDiagram::adjacentExchangeCrossingReduction(list<Vertex_e *> *layer)
{
  /* Scan the verteces in layer from left to right and exchange an
     adjacent pair u, v of verteces, whenever cuv > cvu.
     Repeat until no verteces are exchanged, i.e. the number of
     crossings does not reduce.
  */

  list<Vertex_e *>::iterator iter_l, iter_l_prev;
  int u, v, n, *i = NULL;
  Vertex_e vert;
  bool swap;

  n = layer->size();
  i = new int [n];
  if (!i) return;
  for (u = 0; u < n; u++) i[u] = u;
  do {
    swap = false;
    for (iter_l = layer->begin(), v = u = 0; iter_l != layer->end();
	 iter_l_prev = iter_l++, u = v++) {
      if (iter_l == layer->begin()) continue;
      if (crossing_numbers[i[u]][i[v]] > crossing_numbers[i[v]][i[u]]) {
	vert = **iter_l; **iter_l = **iter_l_prev; **iter_l_prev = vert;
	n = i[u]; i[u] = i[v]; i[v] = n;
	swap = true;
      }
    }
  } while (swap);
  delete i;
}

void
LayerDiagram::splitCrossingReduction(list<Vertex_e *> *layer)
{
  /* If the layer is empty, do nothing.
     Otherwise choose a pivot vertex p in the layer and then place
     all other verteces u in the list Vleft, if cup <= cpu, otherwise
     in the list Vright.
     Apply this algorithm recursively to Vleft and Vright. The output
     is the concatenation of the output of the application to Vleft,
     pivot vertex (!) and the output of the application to Vright.
  */

  int p, u;
  list<Vertex_e *> Vleft, Vright;
  list<Vertex_e *>::iterator iter_l;
  Vertex_e *vp;
  
  if (layer->empty()) return;
  p = layer->size() / 2;
  for (iter_l = layer->begin(), u = 0; iter_l != layer->end();
       iter_l++, u++) {
    if (u == p) { vp = *iter_l; continue; }
    if (crossing_numbers[u][p] <= crossing_numbers[p][u])
      Vleft.push_back(*iter_l);
    else Vright.push_back(*iter_l);
  }
  splitCrossingReduction(&Vleft);
  splitCrossingReduction(&Vright);
  layer->clear();
  for (iter_l = Vleft.begin(); iter_l != Vleft.end(); iter_l++)
    layer->push_back(*iter_l);
  layer->push_back(vp);
  for (iter_l = Vright.begin(); iter_l != Vright.end(); iter_l++)
    layer->push_back(*iter_l);
}

void
LayerDiagram::barycenterCrossingReduction(list<Vertex_e *> *l1, 
					  list<Vertex_e *> *l2)
{
  /* The position x2(u) of the vertex u in layer l2 is avg(u) =
     sum(x1(v)) / deg(u), where v is the vertex in layer l1 and
     (u, v) is an edge, deg(u) is the degree of the vertex u
     (i.e. the number of verteces v).
     If deg(u) is zero, x2(u) is zero.
     If two verteces have the same position, order them arbitrarily.
  */

  list<Vertex_e *>::iterator iter_l1, iter_l2;
  int sum, deg;
  Coord_e_t x;
  
  for (iter_l2 = l2->begin(); iter_l2 != l2->end(); iter_l2++) {
    sum = deg = 0;
    for (iter_l1 = l1->begin(); iter_l1 != l1->end(); iter_l1++)
      if (generationParams->from_maximal) {
	if (isEdge(*iter_l2, *iter_l1)) { sum += (*iter_l1)->x; deg++; }
      }
      else
	if (isEdge(*iter_l1, *iter_l2)) { sum += (*iter_l1)->x; deg++; }
    if (deg) x = sum / deg; else x = 0;
    for (iter_l1 = l2->begin(); iter_l1 != iter_l2; iter_l1++)
      if ((*iter_l1)->x == x) { (x < 0) ? x -= 1 : x += 1; break; }
    (*iter_l2)->x = x; 
  }
}

void
LayerDiagram::medianCrossingReduction(list<Vertex_e *> *l1, 
				      list<Vertex_e *> *l2)
{
  /* The position x2(u) of the vertex u in layer l2 is med(u) =
     x1(v(j/2)), where v(j/2) is the vertex in layer l1 in the middle
     and (u, v) is an edge.
     If the degree of the vertex u is zero (i.e. there are no
     verteces v), x2(u) is zero.
     If two verteces have the same position, the vertex with the odd
     degree goes left, if the parity is the same, order them
     arbitrarily.
  */

  list<Vertex_e *>::iterator iter_l1, iter_l2;
  int deg, i;
  list<Vertex_e *> verts;
  Coord_e_t x;
  
  for (iter_l2 = l2->begin(); iter_l2 != l2->end(); iter_l2++) {
    deg = 0; verts.clear();
    for (iter_l1 = l1->begin(); iter_l1 != l1->end(); iter_l1++)
      if (generationParams->from_maximal) {
	if (isEdge(*iter_l2, *iter_l1)) { verts.push_back(*iter_l1); deg++; }
      }
      else
	if (isEdge(*iter_l1, *iter_l2)) { verts.push_back(*iter_l1); deg++; }
    if (deg) {
      for (iter_l1 = verts.begin(), i = 0; i != (deg / 2); iter_l1++, i++);
      x = (*iter_l1)->x;
    }
    else x = 0;
    for (iter_l1 = l2->begin(); iter_l1 != iter_l2; iter_l1++)
      if ((*iter_l1)->x == x) { 
	((deg / 2) != ((deg + 1) / 2)) ? x -= 1 : x += 1; break;
      }
    (*iter_l2)->x = x;
  }
}

LayerDiagram::LayerDiagram(OrderData *d)
{ 
  if (d) data = d;
  dummy_chains_data = NULL;
  crossing_numbers = NULL;
}

void
LayerDiagram::generateDiagram(Diagram::generationParams_t *params)
{
  list<list<Vertex_e *> *>::iterator iter_ll, iter_ll_prev;
  list<Vertex_e *>::iterator iter_l1, iter_l2;
  Coord_e_t x, y;
  bool the_same;
  const int scale = 2;
  
  if (!params) return;
  generationParams = (generationParams_t *)params;
  // layering
  switch (generationParams->layering) {
  case LAYERING_LONGEST_PATH: longestPathLayering(); break;
  case LAYERING_COFFMAN_GRAHAM: 
    CoffmanGrahamLayering(generationParams->cg_width); break;
  }
  // setup initial vertex coordinates
  for (iter_ll = layers.begin(), y = 0; iter_ll != layers.end();
       iter_ll++) {
    for (iter_l1 = (*iter_ll)->begin(), x = scale * (-(*iter_ll)->size() + 1);
	 iter_l1 != (*iter_ll)->end(); iter_l1++, x += scale * 2) {
      (*iter_l1)->x = x; (*iter_l1)->y = y;
    }
    if (generationParams->from_maximal) y -= scale * 2; else y += scale * 2;
  }
  // insert dummy verteces
  switch (generationParams->dummy_chains_data_type) {
  case DUMMY_D_TREE: dummy_chains_data = new OrderData_Tree(); break;
  case DUMMY_D_LIST: dummy_chains_data = new OrderData_List(); break;
  case DUMMY_D_ARRAY: dummy_chains_data = new OrderData_Array(); break;
  }
  if (!dummy_chains_data) return;
  insertDummyVerteces();
  // crossing reduction
  for (iter_ll = layers.begin(); iter_ll != layers.end();
       iter_ll_prev = iter_ll++) {
    if ((iter_ll == layers.begin()) || ((*iter_ll)->size() < 2) ||
	((*iter_ll_prev)->size() < 2)) continue;
    switch (generationParams->crossing_reduction) {
    case CROSSING_REDUCTION_ADJACENT_EXCHANGE:
      countCrossingNumbers(*iter_ll_prev, *iter_ll);
      adjacentExchangeCrossingReduction(*iter_ll);
      break;
    case CROSSING_REDUCTION_SPLIT:
      countCrossingNumbers(*iter_ll_prev, *iter_ll);
      splitCrossingReduction(*iter_ll);
      break;
    case CROSSING_REDUCTION_BARYCENTER:
      barycenterCrossingReduction(*iter_ll_prev, *iter_ll);
      break;
    case CROSSING_REDUCTION_MEDIAN:
      medianCrossingReduction(*iter_ll_prev, *iter_ll);
      break;
    }
  }
  // adjust vertex coordinates and setup verteces
  for (iter_ll = layers.begin(), y = 0; iter_ll != layers.end();
       iter_ll++) {
    // remove the same vertex positions
    do {
      the_same = false;
      for (iter_l1 = (*iter_ll)->begin(); iter_l1 != (*iter_ll)->end();
	   iter_l1++)
	for (iter_l2 = (*iter_ll)->begin(); iter_l2 != (*iter_ll)->end();
	     iter_l2++)
	  if ((iter_l2 != iter_l1) && (*iter_l2)->x == (*iter_l1)->x) { 
	    ((*iter_l1)->x < 0) ? (*iter_l1)->x -= 1 : (*iter_l1)->x += 1;
	    the_same = true;
	    break;
	  }
    } while (the_same);
    for (iter_l1 = (*iter_ll)->begin(), x = scale * (-(*iter_ll)->size() + 1);
	 iter_l1 != (*iter_ll)->end(); iter_l1++, x += scale * 2) {
      if ((generationParams->crossing_reduction == 
	   CROSSING_REDUCTION_ADJACENT_EXCHANGE) ||
	  (generationParams->crossing_reduction == 
	   CROSSING_REDUCTION_SPLIT)) {
	(*iter_l1)->x = x; (*iter_l1)->y = y;
      }
      verteces.insert(**iter_l1);
    }
    if (generationParams->from_maximal) y -= scale * 2; else y += scale * 2;
  }
}
