/* LatVis
 * Copyright (C) 2002 Jan Outrata <outrataj@phoenix.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "orderdata_array.h"
#include <stdio.h>

bool
OrderData_Array::isElement(ElementId x)
{
  count_t i;

  for (i = 0; i < element_count; i++)
    if (elements[i] == x) return true;
  return false;
}

bool
OrderData_Array::insertPair(ElementId x, ElementId y)
{
  count_t i, j;

  for (i = 0; i < element_count; i++)
    if (elements[i] == x) break;
  for (j = 0; j < element_count; j++)
    if (elements[j] == y) break;
  if ((i == element_count) || (j == element_count)) return false;
  table[i][j] = 1;
  return true;
}

bool
OrderData_Array::erasePair(ElementId x, ElementId y)
{
  count_t i, j;

  for (i = 0; i < element_count; i++)
    if (elements[i] == x) break;
  for (j = 0; j < element_count; j++)
    if (elements[j] == y) break;
  if ((i == element_count) || (j == element_count)) return false;
  if (table[i][j] == 0) return false;
  table[i][j] = 0;
  return true;
}

bool
OrderData_Array::lessTransitive(ElementId x, ElementId y, bool transitivity)
{
  count_t i, j, k;

  for (i = 0; i < element_count; i++)
    if (elements[i] == x) break;
  for (j = 0; j < element_count; j++)
    if (elements[j] == y) break;
  if ((i == element_count) || (j == element_count)) return false;
  if (table[i][j] == 0) return false;
  if (transitivity) return true;
  for (k = 0; k < element_count; k++)
    if ((table[i][k] == 1) && (table[k][j] == 1)) break;
  if (k == element_count) return true;
  return false;
}

void
OrderData_Array::minimal_maximal(set<ElementId> &s, set<ElementId> ss, 
				bool minimal)
{
  /*
    minimal m: does not exist (x, m), x != m, x, m from ss
    maximal M: does not exist (M, x), M != x, M, x from ss
  */
  set<ElementId>::iterator iter_s1, iter_s2;
  
  for (iter_s1 = ss.begin(); iter_s1 != ss.end(); iter_s1++) {
    for (iter_s2 = ss.begin(); iter_s2 != ss.end(); iter_s2++) {
      if (*iter_s1 == *iter_s2) continue;
      if (minimal) { if (lessTransitive(*iter_s2, *iter_s1)) break; }
      else if (lessTransitive(*iter_s1, *iter_s2)) break;
    }
    if (iter_s2 == ss.end()) s.insert(*iter_s1);
  }
}

ElementId
OrderData_Array::least_greatest(set<ElementId> ss, bool least)
{
  /*
    least m: for all x is (m, x), m, x from ss
    greatest M: for all x is (x, M), x, M from ss
  */
  set<ElementId>::iterator iter_s1, iter_s2;
  
  for (iter_s1 = ss.begin(); iter_s1 != ss.end(); iter_s1++) {
    for (iter_s2 = ss.begin(); iter_s2 != ss.end(); iter_s2++) {
      if (*iter_s1 == *iter_s2) continue;
      if (least) { if (!lessTransitive(*iter_s1, *iter_s2)) break; }
      else if (!lessTransitive(*iter_s2, *iter_s1)) break;
    }
    if (iter_s2 == ss.end()) return *iter_s1;
  }
  return NoElement;
}

void
OrderData_Array::lower_upperCone(set<ElementId> &s, set<ElementId> ss, 
				bool lower)
{
  /*
    lower: all m such that for all x is (m, x), x from ss
    upper: all M such that for all x is (x, M), x from ss
  */
  count_t i;
  set<ElementId>::iterator iter_se;
  
  for (i = 0; i < element_count; i++) {
    for (iter_se = ss.begin(); iter_se != ss.end(); iter_se++) {
      if (elements[i] == *iter_se) continue;
      if (lower) { 
	if (!lessTransitive(elements[i], *iter_se)) break;
      }
      else if (!lessTransitive(*iter_se, elements[i])) break;
    }
    if (iter_se == ss.end()) s.insert(elements[i]);
  }
}

void
OrderData_Array::getPaths_(list<set<ElementId> > &l, set<ElementId> s,
			   count_t i, count_t j)
{
  count_t k, m;
  bool transitive = false;
  pair<set<ElementId>::iterator, bool> p;

  for (k = 0; k < element_count; k++)
    if ((table[i][k] == 1) && (table[k][j] == 1)) {
      // element between i, j
      transitive = true;
      for (m = 0; m < element_count; m++)
	if ((table[i][m] == 1) && (table[m][k] == 1)) break;
      if (m != element_count) continue;
      // no element between i, k
      p = s.insert(elements[k]);
      getPaths_(l, s, k, j);
      s.erase(p.first);
    }
  if (!transitive) l.push_back(s);
}

OrderData_Array::~OrderData_Array()
{
  clear();
}

OrderData *
OrderData_Array::getCopy()
{
  OrderData_Array *copy = NULL;
  count_t i, j;

  copy = new OrderData_Array();
  if (!copy) return NULL;
  copy->element_count = element_count;
  copy->elements = new ElementId [element_count];
  for (i = 0; i < element_count; i++) copy->elements[i] = elements[i];
  copy->table = new table_t * [element_count];
  for (i = 0; i < element_count; i++) {
    copy->table[i] = new table_t [element_count];
    for (j = 0; j < element_count; j++) copy->table[i][j] = table[i][j];
  }
  return copy;
}

void
OrderData_Array::clear()
{
  count_t i;
  
  for (i = 0; i < element_count; i++)
    if (table[i] != NULL) delete [] table[i];
  if (table != NULL) { delete [] table; table = NULL; }
  if (elements != NULL) { delete [] elements; elements = NULL; }
  element_count = 0;
}

bool
OrderData_Array::newElement(ElementId x)
{
  count_t new_element_count, i, j;
  ElementId *new_elements;
  table_t **new_table;
  
  if (isElement(x)) return false;
  new_element_count = element_count + 1;
  new_elements = new ElementId [new_element_count];
  for (i = 0; i < element_count; i++) new_elements[i] = elements[i];
  new_elements[i] = x;
  if (elements != NULL) delete [] elements;
  new_table = new table_t * [new_element_count];
  for (i = 0; i < element_count; i++) {
    new_table[i] = new table_t [new_element_count];
    for (j = 0; j < element_count; j++) new_table[i][j] = table[i][j];
    new_table[i][j] = 0;
    if (table[i] != NULL) delete [] table[i];
  }
  new_table[i] = new table_t [new_element_count];
  for (j = 0; j < new_element_count; j++) new_table[i][j] = 0;
  if (table != NULL) delete [] table;
  element_count = new_element_count;
  elements = new_elements;
  table = new_table;
  return true;
}

bool
OrderData_Array::removeElement(ElementId x)
{
  /*
    remove also all pairs containing x
  */
  count_t new_element_count, i, j;
  ElementId *new_elements;
  table_t **new_table;
  count_t skip_index, add1, add2;
  
  if (!isElement(x)) return false;
  new_element_count = element_count - 1;
  new_elements = new ElementId [new_element_count];
  add1 = 0;
  for (i = 0; i < new_element_count; i++) {
    if ((!add1) && (elements[i] == x)) { add1 = 1; skip_index = i; }
    new_elements[i] = elements[i + add1];
  }
  if (elements != NULL) delete [] elements;
  new_table = new table_t * [new_element_count];
  add1 = 0;
  for (i = 0; i < new_element_count; i++) {
    if ((!add1) && (i == skip_index)) add1 = 1;
    new_table[i] = new table_t [new_element_count];
    add2 = 0;
    for (j = 0; j < new_element_count; j++) {
      if ((!add2) && (j == skip_index)) add2 = 1;
      new_table[i][j] = table[i + add1][j + add2];
    }
    if (table[i] != NULL) delete [] table[i];
  }
  if (table != NULL) delete [] table;
  element_count = new_element_count;
  elements = new_elements;
  table = new_table;
  return true;
}

bool
OrderData_Array::orderElements(ElementId x, ElementId y)
{
  count_t i, j, k;

  if ((!isElement(x)) || (!isElement(y))) return false;
  if (x == y) return false;
  if (lessTransitive(y, x)) return false;
  insertPair(x, y);
  for (i = 0; i < element_count; i++)
    if (elements[i] == x) break;
  for (j = 0; j < element_count; j++)
    if (elements[j] == y) break;
  for (k = 0; k < element_count; k++)
    if ((table[j][k] == 1) && (table[k][i] == 0)) insertPair(x, elements[k]);
  for (k = 0; k < element_count; k++)
    if ((table[k][i] == 1) && (table[j][k] == 0)) insertPair(elements[k], y);
  return true;
}

bool
OrderData_Array::removeOrder(ElementId x, ElementId y)
{
  if (x == y) return false;
  if (!lessTransitive(x, y)) return false;
  if (lessTransitive(x, y, false)) erasePair(x, y);
  return true;
}

bool
OrderData_Array::less(ElementId x, ElementId y, bool transitivity)
{
  if ((!isElement(x)) || (!isElement(y))) return false;
  if (x == y) return true;
  return lessTransitive(x, y, transitivity);
}

void
OrderData_Array::getElements(set<ElementId> &s)
{
  count_t i;
  
  for (i = 0; i < element_count; i++) s.insert(elements[i]);
}

void
OrderData_Array::upperElements(set<ElementId> &s, ElementId x)
{
  count_t i, j;
  
  for (i = 0; i < element_count; i++)
    if (elements[i] == x) break;
  if (i == element_count) return;
  for (j = 0; j < element_count; j++)
    if ((table[i][j] == 1) && lessTransitive(x, elements[j], false)) 
      s.insert(elements[j]);
}

void
OrderData_Array::lowerElements(set<ElementId> &s, ElementId x)
{
  count_t i, j;
  
  for (j = 0; j < element_count; j++)
    if (elements[j] == x) break;
  if (j == element_count) return;
  for (i = 0; i < element_count; i++)
    if ((table[i][j] == 1) && lessTransitive(elements[i], x, false)) 
      s.insert(elements[i]);
}

void
OrderData_Array::minimalElements(set<ElementId> &s, set<ElementId> ss)
{
  minimal_maximal(s, ss, true);
}

void
OrderData_Array::maximalElements(set<ElementId> &s, set<ElementId> ss)
{
  minimal_maximal(s, ss, false);
}

ElementId
OrderData_Array::leastElement(set<ElementId> ss)
{
  return least_greatest(ss, true);
}

ElementId
OrderData_Array::greatestElement(set<ElementId> ss)
{
  return least_greatest(ss, false);
}

void
OrderData_Array::lowerCone(set<ElementId> &s, set<ElementId> ss)
{
  lower_upperCone(s, ss, true);
}

void
OrderData_Array::upperCone(set<ElementId> &s, set<ElementId> ss)
{
  lower_upperCone(s, ss, false);
}

ElementId
OrderData_Array::infimum(set<ElementId> ss)
{
  set<ElementId> s1;
  
  lower_upperCone(s1, ss, true);
  return least_greatest(s1, false);
}

ElementId
OrderData_Array::supremum(set<ElementId> ss)
{
  set<ElementId> s1;

  lower_upperCone(s1, ss, false);
  return least_greatest(s1, true);
}

void
OrderData_Array::getPaths(list<set<ElementId> > &l, ElementId x,
			   ElementId y)
{
  count_t i, j;
  set<ElementId> s;
  
  for (i = 0; i < element_count; i++)
    if (elements[i] == x) break;
  for (j = 0; j < element_count; j++)
    if (elements[j] == y) break;
  if ((i == element_count) || (j == element_count)) return;
  if (table[i][j] == 0) return;
  s.insert(x); s.insert(y);
  getPaths_(l, s, i, j);
}
