/* LatVis
 * Copyright (C) 2002 Jan Outrata <outrataj@phoenix.inf.upol.cz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// template of class of undo-redo

#ifndef __UNDOREDO_H__
#define __UNDOREDO_H__

template<class T>
class UndoRedo {
 protected:
  struct Item {
    T data;
    Item *prev;
    Item *next;
  };
  Item *tail;
  Item *actual;
  bool delete_data;

  void insertLast(T d);
  // inserts d at the end
 public:
  UndoRedo(bool del_data);
  ~UndoRedo();
  void save(T d);
  // deletes everything after actual (including) and saves d at the end
  void removeLast();
  // removes the last item if actual is NULL
  T undo(T d);
  // saves d at the end if actual is NULL and sets actual to tail
  // returns: previous item (to actual) or d (and updates actual)
  T redo(T d);
  // returns: next item (to actual) or d (and updates actual)
  // removes tail if actual == tail
};

template<class T>
void
UndoRedo<T>::insertLast(T d)
{
  Item *p;
  
  p = new Item;
  p->data = d;
  p->next = NULL; p->prev = tail;
  if (tail != NULL) tail->next = p;
  tail = p;
}

template<class T>
UndoRedo<T>::UndoRedo(bool del_data)
{
  tail = actual = NULL;
  delete_data = del_data;
}

template<class T>
UndoRedo<T>::~UndoRedo()
{
  actual = NULL;
  while (tail != NULL) removeLast();
}

template<class T>
void
UndoRedo<T>::save(T d)
{
  Item *stop;
  
  if (actual != NULL) {
    stop = actual->prev;
    actual = NULL;
    while (tail != stop) removeLast();
  }
  insertLast(d);
}

template<class T>
void
UndoRedo<T>::removeLast()
{
  if ((actual == NULL) && (tail != NULL)) {
    actual = tail;
    tail = actual->prev;
    if (tail != NULL) tail->next = actual->next;
    if (delete_data) delete actual->data;
    delete actual;
    actual = NULL;
  }
}

template<class T>
T
UndoRedo<T>::undo(T d)
{
  T result;
  
  result = d;
  if (tail != NULL) {
    if (actual == NULL) { insertLast(d); actual = tail; }
    if (actual->prev != NULL) { 
      actual = actual->prev; result = actual->data;
    }
  }
  return result;
}

template<class T>
T
UndoRedo<T>::redo(T d)
{
  T result;
  bool del_data;

  result = d;
  if ((actual != NULL) && (actual != tail)) {
    actual = actual->next;
    result = actual->data;
    if (actual == tail) { 
      actual = NULL; 
      // can't delete data, because we're returning it!
      del_data = delete_data; delete_data = false; 
      removeLast();
      delete_data = del_data;
    }
  }
  return result;
}

#endif
